from functools import partial
import json
import logging
import os
import sys
from time import sleep
from typing import List, Union
from beartype import beartype
from picsellia.utils import bcolors, print_next_bar
from picsellia.decorators import exception_handler
from picsellia.sdk.connexion import Connexion
from picsellia.sdk.dao import Dao
from picsellia import exceptions as exceptions
from picsellia import pxl_multithreading as mlt
import warnings
from beartype.roar import BeartypeDecorHintPep585DeprecationWarning
warnings.filterwarnings("ignore", category=BeartypeDecorHintPep585DeprecationWarning)
logger = logging.getLogger('picsellia')


class Data(Dao):

    def __init__(self, connexion: Connexion, datalake_id: str, id: str, external_url: str, internal_key: str):
        super().__init__(connexion)
        self.datalake_id = datalake_id
        self.id = id
        self.external_url = external_url
        self.internal_key = internal_key

    def __str__(self,): return "{}Data{} object with id: {}".format(bcolors.GREEN, bcolors.ENDC, self.id)

    @exception_handler
    @beartype
    def add_tags(self, tags: Union[str, List[str]]) -> None:
        """Add some tags to data

        You can give a string, a list of string.

        Examples:
            ```python
                data.add_tags("bicyle")
                data.add_tags(["car", "truck", "plane"])
            ```
        """
        if isinstance(tags, str):
            tags = [tags]

        assert tags != [], "Given tags are empty. They can't be empty"
        data = {
            'tags': tags,
            'data_ids': [self.id]
        }
        self.connexion.post('/sdk/v1/datalake/{}/data/tags'.format(self.datalake_id), data=json.dumps(data))
        logger.info("{} tags added to data (id: {}) in datalake {}.".format(len(tags), self.id, self.datalake_id))

    @exception_handler
    @beartype
    def remove_tags(self, tags: Union[str, List[str]]) -> None:
        """Remove some tags of a data

        You can give a string or a list of string.

        Examples:
            ```python
                data.remove_tags("plane")
                data.remove_tags(["truck", "car"])
            ```
        """
        if isinstance(tags, str):
            tags = [tags]

        assert tags != [], "Given tags are empty. They can't be empty"
        data = {
            'tags': tags,
            'data_ids': [self.id]
        }
        self.connexion.delete('/sdk/v1/datalake/{}/data/tags'.format(self.datalake_id), data=json.dumps(data))
        logger.info("{} tags removed from data (id: {}) in datalake {}.".format(len(tags), self.id, self.datalake_id))

    @exception_handler
    @beartype
    def get_tags(self,) -> List[str]:
        """Retrieve the tags of your data.

        Examples:
            ```python
                tags = data.get_tags()
                assert tags == ["bicyle"]
            ```

        Returns:
            List of tags as strings
        """
        r = self.connexion.get('/sdk/v1/datalake/{}/data/{}/tags'.format(self.datalake_id, self.id)).json()
        return r["tags"]

    @exception_handler
    @beartype
    def delete(self,) -> None:
        """Delete data and remove it from datalake.

        :warning: **DANGER ZONE**: Be very careful here!

        Remove this data from datalake, and all of the picture linked to this data.

        Examples:
            ```python
                data.delete()
            ```
        """
        data = {
            'data_ids': [self.id]
        }
        self.connexion.delete('/sdk/v1/datalake/{}/data/delete'.format(self.datalake_id), data=json.dumps(data))
        logger.info("1 asset (id: {}) deleted from datalake {}.".format(self.id, self.datalake_id))


    @exception_handler
    @beartype
    def download(self, target_path : str = './') -> None:
        """Download

        Examples:
            ```python
                data = clt.get_datalake().fetch_data(1)
                data.download('./pictures/')
            ```

        Arguments:
            target_path (str, optional): Target path where data will be downloaded. Defaults to './'.
        """
        path = os.path.join(target_path, self.external_url)
        if self.connexion.download_some_file(False, self.internal_key, path, False):
            logger.info('{} downloaded successfully'.format(self.external_url))
        else:
            logger.error("Could not download {} file".format(self.internal_key))

class MultiData(Dao):

    def __init__(self, connexion: Connexion, datalake_id: str, data_list: List[Data]):
        super().__init__(connexion)
        self.datalake_id = datalake_id

        if data_list == []:
            raise exceptions.NoDataError("A MultiData can't be empty")

        for data in data_list:
            self.check(data)

        self.data_list = data_list
    
    def __str__(self,): return "{}MultiData{} object, size: {}".format(bcolors.GREEN, bcolors.ENDC,len(self))

    def check(self, v):
        if not isinstance(v, Data):
            raise TypeError(v)

    def __getitem__(self, key) -> Union[Data, 'MultiData']:
        if isinstance(key, slice):
            indices = range(*key.indices(len(self.data_list)))
            data = [self.data_list[i] for i in indices]
            return MultiData(self.connexion, self.datalake_id, data)
        return self.data_list[key]

    def __len__(self): return len(self.data_list)

    def __delitem__(self, i):
        if len(self.data_list) < 2:
            raise exceptions.NoDataError("You can't remove from this list the last data. A MultiData can't be empty")

        del self.data_list[i]

    def __setitem__(self, i, v):
        self.check(v)
        self.data_list[i] = v

    @exception_handler
    @beartype
    def add_tags(self, tags: Union[str, List[str]]) -> None:
        """Add some tags to a bunch of data

        You can give a string, a list of string.

        Examples:
            ```python
                whole_data = datalake.fecth_data()
                whole_data.add_tags("never")
                whole_data.add_tags(["gonna", "give", "you", "up"])
            ```
        """
        if isinstance(tags, str):
            tags = [tags]
        assert tags != [], "Given tags are empty. They can't be empty"
        payload = {
            'tags': tags,
            'data_ids': [data.id for data in self.data_list]
        }
        self.connexion.post('/sdk/v1/datalake/{}/data/tags'.format(self.datalake_id), data=json.dumps(payload))
        logger.info("{} tags added to {} data in datalake {}."
                    .format(len(tags), len(self.data_list), self.datalake_id))

    @exception_handler
    @beartype
    def remove_tags(self, tags: Union[str, List[str]]) -> None:
        """Remove some tags on a list of data

        You can give a string, a list of string.

        Examples:
            ```python
                whole_data = datalake.fecth_data()
                whole_data.remove_tags("gonna")
                whole_data.remove_tags(["you"])
            ```
        """
        if isinstance(tags, str):
            tags = [tags]

        assert tags != [], "Given tags are empty. They can't be empty"
        data = {
            'tags': tags,
            'data_ids': [data.id for data in self.data_list]
        }
        self.connexion.delete('/sdk/v1/datalake/{}/data/tags'.format(self.datalake_id), data=json.dumps(data))
        logger.info("{} tags removed from {} data in datalake {}."
                    .format(len(tags), len(self.data_list), self.datalake_id))

    @exception_handler
    @beartype
    def delete(self,) -> None:
        """Delete a bunch of data and remove them from datalake.

        :warning: **DANGER ZONE**: Be very careful here!

        Remove a bunch of data from datalake, and all of the picture linked to all data.

        Examples:
            ```python
                whole_data = datalake.fetch_data(quantity=3)
                whole_data.delete()
            ```
        """
        data = {
            'data_ids': [data.id for data in self.data_list]
        }
        self.connexion.delete('/sdk/v1/datalake/{}/data/delete'.format(self.datalake_id), data=json.dumps(data))
        logger.info("{} data deleted from datalake {}.".format(len(self.data_list), self.datalake_id))

    @exception_handler
    @beartype
    def download(self, target_path: str = './', nb_threads: int = 20) -> None:
        """Download this multi data in given target path


        Examples:
            ```python
                bunch_of_data = client.get_datalake().fetch_data(25)
                bunch_of_data.download('./downloads/')
            ```
        Arguments:
            target_path (str, optional): Target path where to download. Defaults to './'.
            nb_threads (int, optional): Number of threads used to download. Defaults to 20.
        """
        f = partial(mlt.mlt_download_list_data_or_pics, self.connexion, self.datalake_id, target_path)
        mlt.do_multiprocess_things(f, self.data_list, nb_threads)
        logger.info("{} downloaded in {}".format(len(self.data_list), target_path))



from datetime import date
from functools import partial
import json
import logging
import os
from typing import Dict, List, Union
from beartype import beartype
from picsellia.decorators import exception_handler
from picsellia.sdk.connexion import Connexion
from picsellia.sdk.dao import Dao
import picsellia.exceptions as exceptions
from picsellia.utils import bcolors
from picsellia import pxl_multithreading as mlt
import warnings
from beartype.roar import BeartypeDecorHintPep585DeprecationWarning
warnings.filterwarnings("ignore", category=BeartypeDecorHintPep585DeprecationWarning)
logger = logging.getLogger('picsellia')


class Picture(Dao):

    def __init__(self, connexion: Connexion, id: str, external_url: str, internal_key: str, width: int, height: int, tag: list):
        super().__init__(connexion)
        self.id = id
        self.external_url = external_url
        self.internal_key = internal_key
        self.width = width
        self.height = height
        self.tags = tag

    @exception_handler
    @beartype
    def add_annotation(self, data: List[dict], creation_date: date = None, duration: float = None,
                       reviewed: bool = None, is_accepted: bool = None, is_skipped: bool = None,
                       nb_instances: int = None) -> None:
        """Add annotation to this picture.

        Data of annotation is mandatory and its type will be checked here before sending it to platform.
        Other information will be stored in Picsellia database.

        Examples:
            ```
                data = [{
                    "type":"rectangle",
                    "label":"car",
                    "rectangle":{
                        "top":45,
                        "left":120,
                        "width":50,
                        "height":50
                    }
                }]
                pic.add_annotation(data)
            ```
        Arguments:
            data (List[dict]): List of annotations.
            creation_date (date, optional): Creation date; Shall be a 'date' type. Defaults to None.
            duration (float, optional): Duration. Defaults to None.
            reviewed (bool, optional): Reviewed. Defaults to None.
            is_accepted (bool, optional): Is accepted. Defaults to None.
            is_skipped (bool, optional): Is skipped. Defaults to None.
            nb_instances (int, optional): Number of instances. Defaults to None.

        Raises:
            exceptions.TyperError: When annotation is not parsable
        """
        if data == []:
            raise exceptions.TyperError("Data shall not be empty")

        for annot in data:
            if annot == {}:
                raise exceptions.TyperError("An annotation shall not be empty")
            annot_keys = annot.keys()
            if "type" not in annot_keys:
                raise exceptions.TyperError(
                    "'type' key missing from object {}".format(annot))
            supported_types = ["classification", "rectangle", "polygon"]
            if annot["type"] not in supported_types:
                raise exceptions.TyperError(
                    "type must be of {}, found '{}'".format(supported_types, annot["type"]))
            if "label" not in annot_keys:
                raise exceptions.TyperError(
                    "'label' key missing from object {}".format(annot))
            if annot["type"] == "classification":
                pass
            elif annot["type"] == "rectangle":
                if "rectangle" not in annot_keys:
                    raise exceptions.TyperError(
                        "missing 'rectangle' key for object {}".format(annot))
                rect = annot["rectangle"]
                if "top" not in rect.keys():
                    raise exceptions.TyperError(
                        "missing 'top' key in rectangle for object {}".format(annot))
                if "left" not in rect.keys():
                    raise exceptions.TyperError(
                        "missing 'left' key in rectangle for object {}".format(annot))
                if "width" not in rect.keys():
                    raise exceptions.TyperError(
                        "missing 'width' key in rectangle for object {}".format(annot))
                if "height" not in rect.keys():
                    raise exceptions.TyperError(
                        "missing 'height' key in rectangle for object {}".format(annot))
            elif annot["type"] == "polygon":
                if "polygon" not in annot_keys:
                    raise exceptions.TyperError(
                        "missing 'polygon' key for object {}".format(annot))
                poly = annot["polygon"]
                if type(poly) != dict:
                    raise exceptions.TyperError(
                        "'polygon' must be a dict, not {}".format(type(poly)))
                if "geometry" not in poly.keys():
                    raise exceptions.TyperError(
                        "missing 'geometry' key in 'polygon' for object {}".format(annot))
                geometry = poly["geometry"]
                if type(geometry) != list:
                    raise exceptions.TyperError(
                        "'geometry' must be a list, not {}".format(type(geometry)))
                if len(geometry) < 3:
                    raise exceptions.TyperError(
                        "polygons can't have less than 3 points")
                for coords in geometry:
                    if type(coords) != dict:
                        raise exceptions.TyperError(
                            "coordinates in 'geometry' must be a dict, not {}".format(type(coords)))
                    if 'x' not in coords.keys():
                        raise exceptions.TyperError(
                            "missing 'x' coordinate in 'geometry' for object {}".format(annot))
                    if 'y' not in coords.keys():
                        raise exceptions.TyperError(
                            "missing 'y' coordinate in 'geometry' for object {}".format(annot))

        payload = {"data": data}
        if creation_date is not None:
            payload["creation_date"] = creation_date.isoformat()
        if duration is not None:
            payload["duration"] = duration
        if reviewed is not None:
            payload["reviewed"] = reviewed
        if is_accepted is not None:
            payload["is_accepted"] = is_accepted
        if is_skipped is not None:
            payload["is_skipped"] = is_skipped
        if nb_instances is not None:
            payload["nb_instances"] = nb_instances
        self.connexion.put('/sdk/v1/picture/{}/annotations'.format(self.id), data=json.dumps(payload)).json()

    @exception_handler
    @beartype
    def delete_annotations(self,) -> None:
        """Delete all annotations of a picture.

        :warning: **DANGER ZONE**: Be careful here !

        Examples:
            ```python
                pic.delete_annotations()
            ```
        """
        self.connexion.delete('/sdk/v1/picture/{}/annotations'.format(self.id))

    @exception_handler
    @beartype
    def list_annotations(self,) -> List[Dict]:
        """List all annotation of a picture

        Examples:
            ```python
                annotations = pic.list_annotations()
            ```

        Returns:
            A list of dict representing annotations
        """
        r = self.connexion.get('/sdk/v1/picture/{}/annotations'.format(self.id)).json()
        return r["annotations"]

    @exception_handler
    @beartype
    def delete(self,) -> None:
        """Delete picture from its dataset

        :warning: **DANGER ZONE**: Be very careful here!

        Remove this picture and its annotation from the dataset it belongs

        Examples:
            ```python
                pic.delete()
            ```
        """
        self.connexion.delete('/sdk/v1/picture/{}'.format(self.id))
        logger.info("Picture {} removed from dataset".format(self.id))

    @exception_handler
    @beartype
    def download(self, target_path : str = './') -> None:
        """Download this picture into given target path

        Examples:
            ```python
                pic = foo_dataset.get_picture('bar.png')
                pic.download('./pictures/')
            ```

        Arguments:
            target_path (str, optional): Target path where picture will be downloaded. Defaults to './'.
        """
        path = os.path.join(target_path, self.external_url)
        if self.connexion.download_some_file(False, self.internal_key, path, False):
            logger.info('{} downloaded successfully'.format(self.external_url))
        else:
            logger.error("Could not download {} file".format(self.internal_key))


class MultiPicture(Dao):

    def __init__(self, connexion: Connexion, dataset_id: str, picture_list: List[Picture]):
        super().__init__(connexion)
        self.dataset_id = dataset_id

        if picture_list == []:
            raise exceptions.NoDataError("A MultiPicture can't be empty")

        for pic in picture_list:
            self.check(pic)

        self.picture_list = picture_list
    
    def __str__(self,): return "{}MultiPictures{} object, size: {}".format(bcolors.GREEN, bcolors.ENDC,len(self))

    def check(self, v):
        if v is None or not isinstance(v, Picture):
            raise TypeError(v)

    def __getitem__(self, key) -> Union[Picture, 'MultiPicture']:
        if isinstance(key, slice):
            indices = range(*key.indices(len(self.picture_list)))
            pictures = [self.picture_list[i] for i in indices]
            return MultiPicture(self.connexion, self.dataset_id, pictures)
        return self.picture_list[key]

    def __len__(self): return len(self.picture_list)

    def __delitem__(self, i):
        if len(self.picture_list) < 2:
            raise exceptions.NoDataError("You can't remove the last picture. A MultiPicture can't be empty")

        del self.picture_list[i]

    def __setitem__(self, i, v):
        self.check(v)
        self.picture_list[i] = v

    @exception_handler
    @beartype
    def delete(self,) -> None:
        """Delete pictures from their dataset

        :warning: **DANGER ZONE**: Be very careful here!

        Remove this picture and its annotation from the dataset it belongs

        Examples:
            ```python
                pics = dataset.list_pictures()
                pics.delete()
            ```
        """
        payload = {'to_delete': [pic.id for pic in self.picture_list]}
        self.connexion.delete('/sdk/v1/dataset/{}/pictures'.format(self.dataset_id), data=json.dumps(payload))
        logger.info("{} pictures removed from dataset {}".format(len(self.picture_list), self.dataset_id))

    @exception_handler
    @beartype
    def download(self, target_path: str = './', nb_threads: int = 20) -> None:
        """Download this multi picture in given target path


        Examples:
            ```python
                bunch_of_pics = client.get_dataset("foo_dataset").list_pictures()
                bunch_of_pics.download('./downloads/')
            ```
        Arguments:
            target_path (str, optional): Target path where to download. Defaults to './'.
            nb_threads (int, optional): Number of threads used to download. Defaults to 20.
        """
        f = partial(mlt.mlt_download_list_data_or_pics, self.connexion, self.dataset_id, target_path)
        mlt.do_multiprocess_things(f, self.picture_list, nb_threads)
        logger.info("{} downloaded in {}".format(len(self.picture_list), target_path))



import json
import logging
import picsellia.exceptions as exceptions
import os
from uuid import uuid4
import tqdm
import datetime
from picsellia.pxl_multithreading import chunks
import xml.etree.ElementTree as ET
import collections
import warnings
from beartype.roar import BeartypeDecorHintPep585DeprecationWarning
warnings.filterwarnings("ignore", category=BeartypeDecorHintPep585DeprecationWarning)
logger = logging.getLogger('picsellia')


class COCOParser:
    def __init__(self, path: str = None, ):
        self.annpath = path
        if not os.path.isfile(self.annpath):
            raise exceptions.ResourceNotFoundError(
                "Please provide a valid path.")
        with open(self.annpath, 'rb') as f:
            self.data = json.load(f)

    def _find_matching_annotations(self, imgid):
        return [ann for ann in self.data["annotations"] if ann["image_id"] == imgid]

    def _find_category_name(self, catid):
        for cate in self.data["categories"]:
            if cate['id'] == catid:
                return cate['name']
        return False

    def find_dataset_type(self,):
        if self.data['annotations'][0]['bbox'] != [] and self.data['annotations'][0]['segmentation'] == []:
            self.dataset_type = 'detection'
        elif self.data['annotations'][0]['bbox'] != [] and self.data['annotations'][0]['segmentation'] != []:
            self.dataset_type = 'segmentation'
        else:
            self.dataset_type = 'classification'
        return self.dataset_type

    @property
    def images(self,):
        return self.data["images"]

    @property
    def categories(self,):
        return self.data["categories"]

    def generate_images_and_annotations(self, tags=None):
        if tags is None:
            tags = [str(datetime.date.today())]
        logger.debug("Parsing {} ...".format(self.annpath))
        self.find_dataset_type()
        final = []
        for image in tqdm.tqdm(self.data["images"]):
            img_infos = {
                'internal_key': os.path.join(str(uuid4()), image['file_name']),
                'external_url': image['file_name'],
                'height': image["height"],
                'width': image["width"],
                'tags': tags,
                'source': 'sdk'
            }
            ann_list = self._find_matching_annotations(image["id"])
            if self.dataset_type == "detection":
                ann_infos = self._to_picsellia_format_bbox(
                    image["file_name"], ann_list)
            elif self.dataset_type == "segmentation":
                ann_infos = self._to_picsellia_format_poly(
                    image["file_name"], ann_list)
            final.append((
                img_infos,
                ann_infos
            ))
        return final, tags

    def generate_annotations(self, tags=None, rectangle: bool = False):
        if tags is None:
            tags = [str(datetime.date.today())]
        logger.debug("Parsing {} ...".format(self.annpath))
        self.find_dataset_type()
        final = []
        nb_annotations = 0
        nb_objects = 0
        c = collections.Counter()

        for image in tqdm.tqdm(self.data["images"]):
            ann_list = self._find_matching_annotations(image["id"])
            if self.dataset_type == "detection":
                ann_infos = self._to_picsellia_format_bbox(
                    image["file_name"], ann_list)
                final.append(
                    ann_infos
                )
                nb_annotations += 1
                nb_objects += len(ann_infos[1]['data'])
                clss = [e['label'] for e in ann_infos[1]['data']]
                c.update(clss)

            elif self.dataset_type == "segmentation":
                if rectangle and self.data['annotations'][0]['bbox'] != []:
                    ann_infos = self._to_picsellia_format_bbox(
                        image["file_name"], ann_list)
                    final.append(
                        ann_infos
                    )
                else:
                    ann_infos = self._to_picsellia_format_poly(
                        image["file_name"], ann_list)
                    final.append(
                        ann_infos
                    )
                nb_annotations += 1
                nb_objects += len(ann_infos[1]['data'])
                clss = [e['label'] for e in ann_infos[1]['data']]
                c.update(clss)

        return final, dict(c), nb_annotations, nb_objects

    def _to_picsellia_format_bbox(self, imgid, ann_list: list):
        picsellia_list = []
        for ann in ann_list:
            class_name = self._find_category_name(ann['category_id'])
            if not class_name:
                continue
            picsellia_list.append({
                "qa": [],
                "type": "rectangle",
                        "label": class_name,
                        "rectangle": {
                            "top":    ann['bbox'][1],
                            "left":   ann['bbox'][0],
                            "width":  ann['bbox'][2],
                            "height": ann['bbox'][3]
                }

            })

        return [imgid, {'data': picsellia_list}]

    def _to_picsellia_format_poly(self, imgid, ann_list: list):
        """
        TODO -> Create the RLE cases for COCO polygon push.
        """
        picsellia_list = []
        for ann in ann_list:
            class_name = self._find_category_name(ann['category_id'])
            if not class_name:
                continue
            if not self._is_rle(ann):
                tmp = {
                    "qa": [],
                    "type": "polygon",
                            "label": class_name,
                            "polygon": {
                                "geometry": []
                    }
                }
                if not len(ann['segmentation'][0]) % 2 == 0:
                    continue

                indexes = [0]
                if isinstance(ann['segmentation'][0], list):
                    indexes += [(i+2)
                                for i in range(len(ann['segmentation'][0])//2)]
                else:
                    indexes += [(i+2)
                                for i in range(len(ann['segmentation'])//2)]
                for coords in chunks(ann["segmentation"][0], 2):
                    coord = {
                        'x': coords[0],
                        'y': coords[1]
                    }
                    tmp['polygon']['geometry'].append(coord)
                picsellia_list.append(tmp)

        return [imgid, {'data': picsellia_list}]

    def _is_rle(self, annotations):
        return isinstance(annotations['segmentation'], dict)


class PVOCParser:
    def __init__(self, path: str = None):
        self.annpath = path
        if not os.path.isdir(self.annpath):
            raise exceptions.ResourceNotFoundError(
                "Please provide a valid path to your .xml file directory.")
        self.xmlpaths = [os.path.join(self.annpath, e)
                         for e in os.listdir(self.annpath)]

    def _find_dataset_type(self,):
        tree = ET.parse(self.xmlpaths[0])
        root = tree.getroot()
        obj = root.find('object')
        polygon = obj.find('polygon')
        self.dataset_type = "detection" if polygon is None else "segmentation"
        return self.dataset_type

    def _get_class_name(self, obj):
        name = obj.find('name').text.strip().lower()
        return name

    def _get_img_infos(self, root, tags):
        w, h = self._get_width_height(root)
        filename = self._get_file_name(root)
        return {
            'internal_key': os.path.join(str(uuid4()), filename),
            'external_url': filename,
            'height': h,
            'width': w,
            'tags': tags,
            'source': 'sdk'
        }

    def _get_width_height(self, root):
        size = root.find('size')
        width = int(size.find('width').text)
        height = int(size.find('height').text)
        return width, height

    def _get_bbox_coord(self, obj):
        xml_box = obj.find('bndbox')
        xmin = (float(xml_box.find('xmin').text) - 1)
        ymin = (float(xml_box.find('ymin').text) - 1)
        xmax = (float(xml_box.find('xmax').text) - 1)
        ymax = (float(xml_box.find('ymax').text) - 1)

        return {
            'top': ymin,
            'left': xmin,
            'width': xmax-xmin,
            'height': ymax-ymin,
        }

    def _get_file_name(self, root):
        name = root.find('filename').text
        return name

    def _generate_data_bbox(self, tags, root, classes):
        data = []
        for obj in root.iter('object'):
            label = self._get_class_name(obj)
            if len(classes) < 1:
                classes.append((
                    label, 'rectangle'
                ))
            elif (label, 'rectangle') not in classes:
                classes.append((
                    label, 'rectangle'
                ))
            data.append({
                "qa": [],
                "type": "rectangle",
                "label": label,
                "rectangle": self._get_bbox_coord(obj)
            })
        image_infos = self._get_img_infos(root, tags)
        return classes, image_infos, (image_infos['external_url'], {'data': data})

    def _generate_data_poly(self, tags, root, classes):
        data = []
        for obj in root.iter('object'):
            label = self._get_class_name(obj)
            if len(classes) < 1:
                classes.append((
                    label, 'rectangle'
                ))
            elif (label, 'rectangle') not in classes:
                classes.append((
                    label, 'rectangle'
                ))
            data.append({
                "qa": [],
                "type": "polygon",
                "label": label,
                "rectangle": [],
                "polygon": {
                    "geometry": self._get_polygon_coord(obj)
                }
            })
        image_infos = self._get_img_infos(root, tags)
        return classes, image_infos, (image_infos['external_url'], {'data': data})

    def _get_polygon_coord(obj):
        i = 1
        ann = []
        polygon = obj.find('polygon')
        while True:
            x = 'x' + str(i)
            y = 'y' + str(i)
            x1 = polygon.find(x)
            y1 = polygon.find(y)

            if x1 is None or y1 is None:
                break

            x1 = int(x1.text)
            y1 = int(y1.text)
            i += 1
            ann.append({
                'x': x1,
                'y': y1
            })
        return ann

    def _generate_images_labels_annotations(self, tags):

        dataset_type = self._find_dataset_type()
        classes, image_infos, ann_datas = [], [], []

        c = collections.Counter()
        nb_annotations, nb_objects = 0, 0

        dataset_type = self._find_dataset_type()

        for path in self.xmlpaths:
            tree = ET.parse(path)
            root = tree.getroot()

            if dataset_type == "detection":
                classes, image_info, ann_data = self._generate_data_bbox(
                    tags=tags, root=root, classes=classes
                )
            else:
                classes, image_info, ann_data = self._generate_data_poly(
                    tags=tags, root=root, classes=classes
                )
            image_infos.append(image_info)
            ann_datas.append(ann_data)
            nb_annotations += 1
            nb_objects += len(ann_data[1]['data'])
            cc = [e['label'] for e in ann_data[1]['data']]
            c.update(cc)
        return classes, image_infos, ann_datas, dict(c), nb_annotations, nb_objects



import logging
import os
from pathlib import Path
import sys
from typing import Any
from beartype import beartype
import requests
import json
from picsellia.decorators import exception_handler, retry
from picsellia.utils import check_status_code, print_line_return, print_next_bar
import picsellia.exceptions as exceptions
from urllib.parse import urlsplit
import warnings
from beartype.roar import BeartypeDecorHintPep585DeprecationWarning
from urllib.parse import urljoin, urlsplit
from picsellia_connexion_services import TokenServiceConnexion

logger = logging.getLogger('picsellia')
warnings.filterwarnings("ignore", category=BeartypeDecorHintPep585DeprecationWarning)


class Connexion(TokenServiceConnexion):

    def __init__(self, host: str, api_token: str, content_type: str = "application/json") -> None:
        super().__init__(host, api_token, content_type)

    def handle_response(f):
        def decorated(*args, **kwargs):
            try:
                resp = f(*args, **kwargs)
            except Exception:
                raise exceptions.NetworkError(
                    "Server is not responding, please check your host.")
            check_status_code(resp)
            return resp

        return decorated

    @handle_response
    def get(self, path: str, data: dict = None, params: dict = None, stream=False):
        return super().get(path=path, data=data, params=params, stream=stream)

    @handle_response
    def post(self, path: str, data: dict = None, params: dict = None, files: Any = None):
        return super().post(path=path, data=data, params=params, files=files)

    @handle_response
    def put(self, path: str, data: dict = None, params: dict = None):
        return super().put(path=path, data=data, params=params)

    @handle_response
    def patch(self, path: str, data: dict = None, params: dict = None):
        return super().patch(path=path, data=data, params=params)

    @handle_response
    def delete(self, path: str, data: dict = None, params: dict = None):
        return super().delete(path=path, data=data, params=params)

    @exception_handler
    @beartype
    def get_presigned_url(self, method: str, object_name: str, bucket_model: bool = False) -> dict:
        to_send = {
            "method": method,
            "object_name": object_name,
            "bucket_model": bucket_model
        }
        data = json.dumps(to_send)
        url = urljoin(self.host, '/sdk/v1/generate_presigned_url')
        r = requests.post(url, data=data, headers=self.headers)

        if r.status_code != 200:
            raise exceptions.PicselliaError("Errors while getting a presigned url")

        r = r.json()
        if "url" not in r:
            raise exceptions.PicselliaError("Errors while getting a presigned url. Unparsable response")

        return r

    @retry((KeyError, ValueError, exceptions.PicselliaError), total_tries=4)
    @exception_handler
    @beartype
    def push_to_s3(self, path: str, internal_key: str) -> requests.Response:
        response = self.get_presigned_url("post", object_name=internal_key)
        with open(path, 'rb') as file:
            r = requests.post(response["url"], data=response["fields"], files={'file': (internal_key, file)})

        if r.status_code != 204:
            raise exceptions.PicselliaError("Could not retrieve presigned url")

        return r

    @exception_handler
    @beartype
    def download_some_file(self, is_large: bool, object_name: str, path: str, bucket_model: bool = True):

        if is_large:
            return self._download_large_file(object_name, path, bucket_model)
        
        # If file shall not be large
        try:
            return self._download_file(object_name, path, bucket_model)
        except Exception:
            return self._download_large_file(object_name, path, bucket_model)
    
    @exception_handler
    @beartype
    def _download_file(self, object_name: str, target_path: str, bucket_model: bool = True) -> bool:
        """Download a file from S3.

        Arguments:
            object_name (str): Object s3 Key name to download
            target_path (str): Target path of file to download object
        """
        response = self.get_presigned_url('get', object_name, bucket_model=bucket_model)

        parent_path = Path(target_path).parent.absolute()
        if not os.path.exists(parent_path):
            os.makedirs(parent_path)

        with open(target_path, 'wb') as handler:
            response = requests.get(response["url"])
            total_length = response.headers.get('content-length')
            if total_length is None:
                return False

            handler.write(response.content)
            return True

    @exception_handler
    @beartype
    def _download_large_file(self, object_name: str, path: str, bucket_model: bool = True) -> bool:
        """Downloads a large file from the server.

        Arguments:
            object_name (str): [Object s3 Key name to download]
            path (str): [Direction to download it]
        """

        response = self.get_presigned_url('get', object_name, bucket_model=bucket_model)

        parent_path = Path(path).parent.absolute()
        if not os.path.exists(parent_path):
            os.makedirs(parent_path)
            
        sys.stdout.write("-----")
        with open(path, 'wb') as handler:
            response = requests.get(response["url"], stream=True)
            total_length = response.headers.get('content-length')
            if total_length is None:
                return False
            total_length = int(total_length)
            dl = 0
            count = 0
            last_done = -1
            for data in response.iter_content(chunk_size=4096):
                dl += len(data)
                handler.write(data)
                done = int(50 * dl / total_length)
                if done != last_done:
                    print_next_bar(dl, total_length)
                    last_done = done
                count += 1
        sys.stdout.write("--*--")
        print_line_return()
        return True

    @exception_handler
    @beartype
    def _init_multipart(self, object_name: str) -> str:
        """Initialize a multipart push

        Arguments:
            object_name (str): object name to download

        Returns:
            The upload id if everything went well
        """

        to_send = {"object_name": object_name}
        url = urljoin(self.host, '/sdk/v1/init_multipart_upload')

        try:
            r = requests.get(url, data=json.dumps(to_send))
        except Exception:
            raise exceptions.NetworkError("Server is not responding, please check your host.")

        if r.status_code != 200 or "upload_id" not in r.json():
            raise exceptions.NetworkError(
                "Impossible to init a multipart upload because :\n{}".format(r.text))

        r = r.json()
        if "upload_id" not in r:
            raise exceptions.NetworkError("Response when initiating multipart is unparsable")

        return r["upload_id"]

    @exception_handler
    @beartype
    def _get_url_for_part(self, upload_id, no_part, object_name) -> str:
        """Get a pre-signed url to upload a part of Checkpoints or SavedModel
        Raises:
            NetworkError: If it impossible to initialize upload

        """
        to_send = {
            "object_name": object_name,
            "upload_id": upload_id,
            "part_no": no_part
        }
        url = urljoin(self.host, '/sdk/v1/generate_post_part_url')

        try:
            r = requests.post(url, data=json.dumps(to_send))
        except Exception:
            raise exceptions.NetworkError("Impossible to get an url for part")

        if r.status_code != 200:
            raise exceptions.NetworkError("Impossible to get an url.. because :\n{}".format(r.text))

        r = r.json()
        if "url" not in r:
            raise exceptions.NetworkError("Response when getting an url for a part is unparsable")

        return r["url"]

    @exception_handler
    @beartype
    def _upload_parts(self, upload_id, file_path, object_name):
        if not os.path.exists(file_path):
            raise exceptions.PicselliaError("Impossible to upload part in an empty filepath")

        max_size = 5 * 1024 * 1024
        file_size = os.path.getsize(file_path)
        upload_by = int(file_size / max_size) + 1
        sys.stdout.write("-----")
        with open(file_path, 'rb') as f:

            parts = []
            for part in range(1, upload_by + 1):
                url = self._get_url_for_part(upload_id, part, object_name)
                try:
                    file_data = f.read(max_size)
                except Exception:
                    raise exceptions.NetworkError("Impossible to put part no {}".format(part))

                try:
                    res = requests.put(url, data=file_data)
                except Exception:
                    raise exceptions.NetworkError("Impossible to get an url for part")

                if res.status_code != 200:
                    raise exceptions.NetworkError(
                        "Impossible to put part no {}\n because {}".format(part, res.text))
                etag = res.headers['ETag']
                parts.append({'ETag': etag, 'PartNumber': part})
                print_next_bar(part, upload_by)
        sys.stdout.write("--*--")
        print_line_return()
        return parts

    @exception_handler
    @beartype
    def _complete_part_upload(self, upload_id, parts, object_name):
        """
            Complete the upload a part of Checkpoints or SavedModel
            Raises:
                NetworkError: If it impossible to initialize upload
        """
        to_send = {
            "object_name": object_name,
            "upload_id": upload_id,
            "parts": parts,
        }
        url = urljoin(self.host, '/sdk/v1/complete_part_upload')

        try:
            r = requests.post(url, data=json.dumps(to_send), headers=self.headers)
        except Exception:
            raise exceptions.NetworkError("Impossible to get an url..")

        if r.status_code != 201:
            raise exceptions.NetworkError("Impossible to get an url.. because :\n{}".format(r.text))

        return r

    @exception_handler
    @beartype
    def _send_large_file(self, path: str = None, object_name: str = None):
        """Uploads a large file to the experiment.

        Arguments:
            path (str, optional): [Path to the file to upload]. Defaults to None.
            name (str, optional): [Name of the file]. Defaults to None.
            object_name (str, optional): [Key name for S3 storage]. Defaults to None.
        """
        upload_id = self._init_multipart(object_name)
        parts = self._upload_parts(upload_id, path, object_name)
        response = self._complete_part_upload(upload_id, parts, object_name)
        return response

    @exception_handler
    @beartype
    def _send_file(self, path: str, object_name: str):
        """Uploads a small file to the experiment.

        Arguments:
            path (str, optional): [Path to the file to upload]
            name (str, optional): [Name of the file]
            object_name (str, optional): [Key name for S3 storage]
            network_id (str, optional): [ID of the network to attach file]
        """
        if not os.path.isfile(path):
            raise FileNotFoundError("{} not found".format(path))

        response = self.get_presigned_url(method='post', object_name=object_name, bucket_model=True)

        with open(path, 'rb') as f:
            files = {'file': (path, f)}
            try:
                http_response = requests.post(response['url'], data=response['fields'], files=files)
            except Exception:
                raise exceptions.NetworkError("Impossible to get an url for part")

        return http_response

    @exception_handler
    @beartype
    def upload_file(self, path: str, object_name: str) -> requests.Response:
        """Upload a single file to the server.
        If file is bigger than 5Mb, it will send it by multipart.

        Arguments:
            path (str): [Absolute path to the file]
            object_name (str): [Bucket prefix s3]

        Raises:
            FileNotFoundError: [File does not exists]
            exceptions.NetworkError: [Platform does not respond]

        Returns:
            A requests.Response of the send file http request
        """
        if not os.path.isfile(path):
            raise FileNotFoundError("{} not found".format(path))

        filesize = Path(path).stat().st_size
        if filesize <= 5*1024*1024:
            response = self._send_file(path, object_name)
            if response.status_code != 204:
                raise exceptions.PicselliaError("Could not upload file : {}".format(response.status_code))
        else:
            upload_id = self._init_multipart(object_name)
            parts = self._upload_parts(upload_id, path, object_name)
            response = self._complete_part_upload(upload_id, parts, object_name)

        return response



import base64
from datetime import datetime, timezone
import enum
import json
import logging
from re import I
from typing import Dict, List, Union
from picsellia.exceptions import ImpossibleAction
from picsellia.sdk.connexion import Connexion
from picsellia.sdk.dataset import Dataset
from picsellia.sdk.project import Project
from picsellia.sdk.model import Model
from picsellia.sdk.dao import Dao
from picsellia.sdk.model import Model
from picsellia.types.prediction_format import PredictionFormat
import os
import warnings
from beartype.roar import BeartypeDecorHintPep585DeprecationWarning
from picsellia_connexion_services import JwtServiceConnexion

logger = logging.getLogger('picsellia')
warnings.filterwarnings("ignore", category=BeartypeDecorHintPep585DeprecationWarning)

class ServiceMetrics(str, enum.Enum):

    def __new__(cls, *args, **kargs):
        value = len(cls.__members__) + 1
        obj = str.__new__(cls)
        obj._value_ = value
        return obj

    def __init__(self, service, is_aggregation):
        self.service = service
        self.is_aggregation = is_aggregation
        return

    PREDICTIONS_OUTLYING_SCORE = 'ae_outlier', False
    PREDICTIONS_DATA = 'metrics', False
    REVIEWS_OBJECT_DETECTION_STATS = 'object_detection', False
    REVIEWS_CLASSIFICATION_STATS = 'classification', False
    REVIEWS_LABEL_DISTRIBUTION_STATS = 'label_distribution', False

    AGGREGATED_LABEL_DISTRIBUTION = 'label_distribution', True
    AGGREGATED_OBJECT_DETECTION_STATS = 'object_detection', True
    AGGREGATED_PREDICTIONS_DATA = 'metrics', True
    AGGREGATED_DRIFTING_PREDICTIONS = 'ks_drift', True


class Deployment(Dao):

    def __init__(self, connexion: Connexion, id: str, name: str, oracle_host : str = None, serving_host : str = None):
        super().__init__(connexion)
        self.id = id
        self.name = name

        if oracle_host != None:
            self.oracle_connexion = JwtServiceConnexion(oracle_host, {"api_token": self.connexion.api_token, "deployment_id": id})

        if serving_host != None:
            self.serving_connexion = JwtServiceConnexion(serving_host, {"api_token": self.connexion.api_token, "deployment_id": id})

    def retrieve_information(self) -> dict:
        """Retrieve some information about this deployment from service.

        Examples:
            ```python
                print(my_deployment.retrieve_information())
                >>> {

                }
            ```
        """
        assert self.oracle_connexion != None, 'This action cannot be done'

        return self.oracle_connexion.get(path='/deployment/{}'.format(self.id)).json()

    def delete(self,) -> None:
        self.connexion.delete('/sdk/v1/deployment/{}/delete'.format(self.id))
        logging.info("Deployment {} deleted.".format(self.name))
        return None

    def predict(self, fpath: str) -> dict:
        """_summary_

        Examples:
            ```python
                deployment = client.get_deployment(
                    name="awesome-deploy"
                )
                deployment.predict()
            ```
        Arguments:
            fpath (str): path to the image to predict

        Returns:
            (dict): _description_
        """        
        assert self.serving_connexion != None, 'This action cannot be done'

        files = {'media': open(fpath, 'rb')}
        return self.serving_connexion.post(path='/predict/{}'.format(self.id), files=files).json()

    def init_feedback_loop(self, dataset: Dataset) -> None:
        """Initialize the Feedback Loop for a Deployment.
        This way, you will be able to attached reviewed predictions to the Dataset.
        This is a great option to increase your training set with quality data.

        Examples:
            ```python
                dataset = client.get_dataset(
                    name="my-dataset",
                    version="latest"
                )
                deployment = client.get_deployment(
                    name="awesome-deploy"
                )
                deployment.init_feedback_loop(
                    dataset
                )
            ```
        Arguments:
            dataset (Dataset): a connected (Dataset)
        """        
        data = {
            "dataset_id": dataset.id,
        }
        self.connexion.post('/sdk/v1/deployment/{}/pipeline/feedback_loop'.format(self.id), data=json.dumps(data))
        logger.info("Feedback loop setup for deployment {}\nNow you will be able to add predictions\nto dataset: {}/{}".format(self.name, dataset.name, dataset.version))
        return

    def update_feedback_loop(self, dataset: Dataset) -> None:
        """Update the Feedback Loop for a Deployment.
        This way, you will be able to attached reviewed predictions to the updated Dataset.
        This is a great option to increase your training set with quality data.

        Examples:
            ```python
                dataset = client.get_dataset(
                    name="my-dataset",
                    version="new-version"
                )
                deployment = client.get_deployment(
                    name="awesome-deploy"
                )
                deployment.update_feedback_loop(
                    dataset
                )
            ```
        Arguments:
            dataset (Dataset): a connected (Dataset)
        """        
        data = {
            "dataset_id": dataset.id,
            "disable": False
        }
        self.connexion.post('/sdk/v1/deployment/{}/pipeline/feedback_loop/edit'.format(self.id), data=json.dumps(data))
        logger.info("Feedback loop setup for deployment {}".format(self.name))
        logger.info("Now you will be able to add predictions to dataset: {}/{}".format(dataset.name, dataset.version))
        return

    
    def init_continuous_training(self, project: Project, dataset: Dataset, model: Model, 
                                    parameters: dict, training_type: str = None, 
                                        scan_config : dict = None, policy: str = None, 
                                                feedback_loop_trigger: int = None) -> None:
        """Initiliaze and activate the continuous features of picsellia. 🥑
           A Training will be triggered using the configured Dataset
           and Model as base whenever your Deployment pipeline hit the trigger.

        Examples:
            # Let's initialize a continuous training pipeline that will be trigger 
            # every 150 new predictions reviewed by your team. 
            # We will use the same training parameters as those used when building the first model.

            ```python
                project = client.get_project(name="my-project")
                dataset = project.get_dataset(name="my-dataset", version="latest")
                model = client.get_model(name="my-model")
                deployment = client.get_deployment("awesome-deploy")
                experiment = model.get_source_experiment()
                parameters = experiment.get_log('parameters')
                training_type = "experiment"
                feedback_loop_trigger = 150
                deployment.init_continuous_training(
                    project, dataset, model,
                    parameters, training_type, 
                    feedback_loop_trigger=feedback_loop_trigger
                )
            ```

        Arguments:
            project (Project): The project that will host your pipeline. 
            dataset (Dataset): The Dataset that will be used as training data for your training.
            model (Model): The exported Model to perform transfert learning from.
            parameters (dict): Training parameters.
            training_type (str, optional): Either `experiment` or `scan`.
            scan_config (dict, optional): Scan configuration dict. [more info](https://doc.picsellia.com/docs/initialize-a-scan)
            policy (str, optional): Early Stopping policy.
            feedback_loop_trigger (int, optional): Number of images that need to be review to trigger the training.
        """        
        data = {
            "project_id": project.id,
            "dataset_id": dataset.id,
            "model_id": model.id,
            "parameters": parameters,
            "training_type": training_type,
            "scan_config": scan_config,
            "policy": policy,
            "fl_trigger": feedback_loop_trigger
        }
        
        self.connexion.post('/sdk/v1/deployment/{}/pipeline/continuous_training'.format(self.id), data=json.dumps(data))
        logger.info("Continuous training setup for deployment {}\nNow you will be able to add predictions\nto dataset: {}/{}".format(self.name, dataset.name, dataset.version))
        return

    def activate_continuous_training(self,):
        """Activate your continuous training pipeline.

        Examples:
            ```python
                deployment = client.get_deployment("awesome-deploy")
                deployment.activate_continuous_training()
            ```
        """
        data = {
            "active": True
        }
        self.connexion.post('/sdk/v1/deployment/{}/pipeline/continuous_training/toggle'.format(self.id), data=json.dumps(data))
        logger.info("Continuous training for deployment {} activated.".format(self.name))
        return

    def deactivate_continuous_training(self,):
        """Deactivate your continuous training pipeline.

        Examples:
            ```python
                deployment = client.get_deployment("awesome-deploy")
                deployment.deactivate_continuous_training()
            ```
        """
        data = {
            "active": False
        }   
        self.connexion.post('/sdk/v1/deployment/{}/pipeline/continuous_training/toggle'.format(self.id), data=json.dumps(data))
        logger.info("Continuous training for deployment {} deactivated.".format(self.name))
        return

    def get_stats(self, service : ServiceMetrics, model: Model = None, from_timestamp: float = None, to_timestamp: float = None,
                  since: int = None, includes : List[str] = None, excludes : List[str] = None, tags : str = None) -> dict:
        """Retrieve stats of this deployment stored in Picsellia environment.

        Mandatory param is "service" an enum of type ServiceMetrics. Values possibles are : 
            PREDICTIONS_OUTLYING_SCORE
            PREDICTIONS_DATA
            REVIEWS_OBJECT_DETECTION_STATS
            REVIEWS_CLASSIFICATION_STATS
            REVIEWS_LABEL_DISTRIBUTION_STATS

            AGGREGATED_LABEL_DISTRIBUTION
            AGGREGATED_OBJECT_DETECTION_STATS
            AGGREGATED_PREDICTIONS_DATA
            AGGREGATED_DRIFTING_PREDICTIONS

        For aggregation, computation may not have been done by the past.
        You will need to force computation of these aggregations and retrieve them again.


        Examples:
            ```python
                my_deployment.get_stats(ServiceMetrics.PREDICTIONS_DATA)
                my_deployment.get_stats(ServiceMetrics.AGGREGATED_DRIFTING_PREDICTIONS, since=3600)
                my_deployment.get_stats(ServiceMetrics.AGGREGATED_LABEL_DISTRIBUTION, model_id=1239012)

            ```
        Arguments:
            service (str): service queried
            model (Model, optional): Model that shall be used when retrieving data. Defaults to None.
            from_timestamp (float, optional): System will only retrieve prediction data after this timestamp. Defaults to None.
            to_timestamp (float, optional): System will only retrieve prediction data before this timestamp. Defaults to None.
            since (int, optional): System will only retrieve prediction data that are in the last seconds given by this value. Defaults to None.
            includes (List[str], optional): Research will includes these ids and excludes others. Defaults to None.
            excludes (List[str], optional): Research will excludes these ids. Defaults to None.
            tags (str, optional): Research will be done filtering by tags. Defaults to None.
                                  tags need to be parsable like "tag1:value,tag2:value2"

        Returns:
            A dict with queried statistics about the service you asked
        """
        assert self.oracle_connexion != None, "You can't use this function with this deployment. Please contact the support."

        filter = self._build_filter(service=service.service, model=model, from_timestamp=from_timestamp, to_timestamp=to_timestamp,
                                    since=since, includes=includes, excludes=excludes, tags=tags)

        if service.is_aggregation:
            resp = self.oracle_connexion.get(path='/deployment/{}/stats'.format(self.id), data=filter).json()
            if "infos" in resp and "info" in resp["infos"]:
                logging.info('This computation is outdated or has never been done.\n" \
                             You can compute it again by calling launch_computation with exactly the same params.')
            return resp
        else:
            return self.oracle_connexion.get(path='/deployment/{}/predictions/stats'.format(self.id), data=filter).json()


    def _build_filter(self, service : str, model : Model = None, from_timestamp: float = None, to_timestamp: float = None,
                  since: int = None, includes : List[str] = None, excludes : List[str] = None, tags : str = None) -> dict:
        
        filter = { "service" : service }

        if model != None:
            filter["model_id"] = model.id

        if from_timestamp != None:
            filter["from_timestamp"] = from_timestamp

        if to_timestamp != None:
            filter["to_timestamp"] = to_timestamp

        if since != None:
            filter["since"] = since

        if includes != None:
            filter["includes"] = includes

        if excludes != None:
            filter["excludes"] = excludes

        if tags != None:
            filter["tags"] = tags

        return filter


    def monitor(self, model : Model, image_path : str, latency : float, picture_id : str, filename: str, height : int, width: int,
                source : str, tags: Dict[str, Union[str, int, float]], prediction : PredictionFormat,
                shadow_model : Model = None, shadow_latency : float = None, shadow_raw_predictions : PredictionFormat = None):
        with open(image_path, "rb") as img_file:
            encoded_image = base64.b64encode(img_file.read()).decode('utf-8')

        data = {
            "picture_id": picture_id,
            "filename": filename,
            "raw_predictions": prediction.to_payload(),
            "latency": latency,
            'height': height,
            'width': width,
            'source': source,
            'model_type': model.type,
            "model": model.id,
            "timestamp": str(datetime.now(timezone.utc).timestamp()),
            "tags": tags,
            "image": encoded_image
        }

        if shadow_model != None:
            if shadow_latency == None or shadow_raw_predictions == None:
                raise ImpossibleAction('Shadow latency and shadow raw predictions shall be defined if you want to push a shadow model result')
            data["shadow_model"] = shadow_model.id
            data["shadow_latency"] = shadow_latency
            data["shadow_raw_predictions"] = shadow_raw_predictions


        return self.oracle_connexion.post(path='/deployment/{}/add'.format(self.id), data=data)



from beartype import beartype
from picsellia.decorators import exception_handler
from picsellia.sdk.connexion import Connexion
from picsellia.sdk.dao import Dao
from picsellia.sdk.run import Run
import warnings
from beartype.roar import BeartypeDecorHintPep585DeprecationWarning
warnings.filterwarnings("ignore", category=BeartypeDecorHintPep585DeprecationWarning)

class Scan(Dao):

    def __init__(self, connexion: Connexion, project_token: str, id: str, name: str) -> None:
        super().__init__(connexion)
        self.project_token = project_token
        self.id = id
        self.name = name

    @exception_handler
    @beartype
    def launch(self,) -> None:
        """Distribute runs remotely for this scan.

        :information-source: The remote environment has to be setup prior launching the experiment.
        It defaults to our remote training engine.

        Examples:
            ```python
                scan.launch()
            ```
        """
        self.connexion.post('/sdk/v1/scan/{}/launch'.format(self.id))

    @exception_handler
    @beartype
    def _convert_response_to_run(self, response: dict) -> Run:
        script = None
        script_object_name = None
        if "script" in response.keys():
            script = response["script"]
            script_object_name = response["script_object_name"]
        return Run(
            self.connexion,
            response["id"],
            response["config"],
            script,
            script_object_name,
            response["requirements"]
        )

    @exception_handler
    @beartype
    def get_run_by_id(self, id: str) -> Run:
        """Retrieve a run object by its id.

        Examples:
            ```python
                scan.get_run_by_id("cb750009-4e09-42bb-8c84-cc78aa004bf0")
            ```
        Arguments:
            id (str): id (primary key) of the run on Picsellia

        Returns:
            A (Run) object manipulable
        """
        r = self.connexion.get('/sdk/v1/run/{}'.format(id)).json()
        run = self._convert_response_to_run(r["run"])
        return run

    @exception_handler
    @beartype
    def get_next_run(self,) -> Run:
        """Get next available Run for Scan.

        Examples:
            ```python
                scan.get_next_run()
            ```

        Returns:
            A (Run) object manipulable
        """
        r = self.connexion.get('/sdk/v1/scan/{}/run/next'.format(self.id)).json()
        run = self._convert_response_to_run(r["run"])
        return run





import abc


class PredictionFormat(abc.ABC):

    def __init__(self):
        pass
        
    @abc.abstractmethod
    def check_validity(self,) -> bool:
        pass
    
    @abc.abstractmethod
    def to_payload(self,) -> dict:
        pass

    


import logging
from functools import partial
from typing import List, Union
import picsellia.pxl_multithreading as mlt
from beartype import beartype
from picsellia import exceptions
from picsellia.decorators import exception_handler
from picsellia.exceptions import InvalidQueryError, NoDataError, PicselliaError
from picsellia.sdk.connexion import Connexion
from picsellia.sdk.dao import Dao
from picsellia.sdk.data import Data, MultiData

from picsellia.utils import print_next_bar
import warnings
from beartype.roar import BeartypeDecorHintPep585DeprecationWarning
warnings.filterwarnings("ignore", category=BeartypeDecorHintPep585DeprecationWarning)
logger = logging.getLogger('picsellia')


class Datalake(Dao):

    def __init__(self, connexion: Connexion, id: str):
        super().__init__(connexion)
        self.id = id

    @exception_handler
    @beartype
    def get_resource_url_on_platform(self,) -> str:
        """Get platform url of this resource.

        Examples:
            ```python
                print(foo_dataset.get_resource_url_on_platform())
                >>> https://app.picsellia.com/datalake/62cffb84-b92c-450c-bc37-8c4dd4d0f590
            ```

        Returns:
            Url on Platform for this resource
        """

        return "{}/datalake/{}".format(self.connexion.host, self.id)

    @exception_handler
    @beartype
    def _convert_response_to_data(self, response):
        return Data(
            self.connexion,
            self.id,
            response["picture_id"],
            response["external_url"],
            response["internal_key"]
        )

    @exception_handler
    @beartype
    def upload_data(self, filepaths: Union[str, List[str]],
                    tags: List[str] = [], source: str = 'sdk', nb_threads: int = 20) -> Union[Data, MultiData]:
        """Upload data into this datalake.

        Upload files of pictures, representing data, into a datalake.
        You can give some tags as a list.
        You can give a source for your data.

        Examples:
            ```python
                lake = client.get_datalake()
                lake.upload_data(filepaths=["pics/twingo.png", "pics/ferrari.png"], tags=["car"], source='camera_one')
                lake.upload_data(filepaths="pics/truck.png", tags=["megacar"], source='camera_two')
            ```
        Arguments:
            filepaths (str or List[str]): Filepath of your data
            tags (List[str], optional): Tags that must be given to data. Defaults to [].
            source (str, optional): Source of your data. Defaults to 'sdk'.

        Returns:
            A (Data) object or a (MultiData) object that wraps a list of Data.
            You can manipulate this object returned and add some tags or feeding a Dataset.
        """
        if isinstance(filepaths, str):
            filepaths = [filepaths]

        logger.info("🌎 Starting upload ..") 

        f = partial(mlt.mlt_upload_data, self.connexion, self.id, tags, source)
        shared_data_list = mlt.do_multiprocess_things(f, filepaths, nb_threads)


        if len(shared_data_list) != len(filepaths):
            logger.error("❌ {} data not uploaded. Check readability / check paths".format(len(filepaths) - len(shared_data_list)))

        if len(shared_data_list) == 0:
            raise PicselliaError("Nothing has been uploaded. The image may not have been readable \
                                  or something went wrong while contacting picsellia.")
        elif len(shared_data_list) == 1:
            logger.info("✅ {} data uploaded.".format(shared_data_list[0].external_url))
            return shared_data_list[0]
        else:
            logger.info("✅ {} data uploaded.".format(len(shared_data_list)))
            return MultiData(self.connexion, self.id, shared_data_list)

    @exception_handler
    @beartype
    def list_data(self) -> MultiData:
        """List all pictures of your organization.

        List all pictures in this datalake.
        If there is no data, raise a NoDataError exception.

        Returned object is a MultiData. An object that allows manipulation of a bunch of data.
        You can add tags on them or feed a dataset with them.

        Examples:
            ```python
                lake = client.get_datalake()
                data = lake.list_data()
            ```

        Raises:
            NoDataError: When datalake has no data, raise this exception.

        Returns:
            A (MultiData) object that wraps a list of (Data).
            You can manipulate this object returned and add some tags or feeding a dataset.
        """
        r = self.connexion.get('/sdk/v1/datalake/{}/data'.format(self.id)).json()
        list_data = list(map(self._convert_response_to_data, r["data"]))

        if len(list_data) == 0:
            raise NoDataError("No assets found in this datalake")

        return MultiData(self.connexion, self.id, list_data)

    @exception_handler
    @beartype
    def fetch_data(self, quantity: int = -1, tags: Union[str, List[str]] = [], no_tags: bool=False) -> MultiData:
        """Fetch data of datalake

        Use this method if you want a certain amount of data.
        You can precise `quantity` : it will limit the number of data retrieved.
        You can precise some `tags` : retrieved data SHALL HAVE all these tags.

        Examples:
            ```python
                lake = client.get_datalake()
                data = lake.fetch_data(quantity=10, tags="car")
                assert len(data) == 2
            ```
            Notes
            - We uploaded 2 pictures with tag "car" previously
            - Return type is MultiData but behaves like a list: you can slice it (-> MultiData) or get an item (-> Data)

        Arguments:
            quantity (int, optional): Number of data max to retrieve. Defaults to 1.
            tags (str or List[str], optional): List of tags that data shall have. Defaults to [].

        Raises:
            NoDataError: When there is no data to retrieve, raise this exception.

        Returns:
            A (MultiData) object that wraps a list of Data.
            You can manipulate this object returned and add some tags or feeding a dataset.
        """
        if no_tags and tags != []:
            raise InvalidQueryError("Please set no_tags to False if you want to use tags.")
        if isinstance(tags, str):
            tags = [tags]

        params = {
            'quantity': quantity,
            'no_tags': no_tags
        }
        if tags != []:
            params["tags"] = ",".join(tags)
        r = self.connexion.get('/sdk/v1/datalake/{}/data'.format(self.id), params=params).json()
        list_data = list(map(self._convert_response_to_data, r["data"]))

        if len(list_data) == 0:
            raise NoDataError("No assets found with tags {}".format(tags))

        return MultiData(self.connexion, self.id, list_data)



from functools import partial
import json
import logging
import os
from typing import List, Union
from beartype import beartype
from picsellia.exceptions import PicselliaError
import picsellia.pxl_multithreading as mlt
import picsellia
from picsellia.decorators import exception_handler
from picsellia.sdk.connexion import Connexion
from picsellia.sdk.data import Data, MultiData
from picsellia.sdk.datalake import Datalake
from picsellia.sdk.dataset import Dataset
from picsellia.sdk.deployment import Deployment
from picsellia.sdk.model import Model
from picsellia.sdk.organization import Organization
from picsellia.sdk.project import Project
from picsellia.utils import bcolors
import warnings
from beartype.roar import BeartypeDecorHintPep585DeprecationWarning
warnings.filterwarnings("ignore", category=BeartypeDecorHintPep585DeprecationWarning)
logger = logging.getLogger('picsellia')


class Client:
    """
    Picsellia SDK Client shall be used to communicate with Picsellia services.

    You need an API Token, available on web platform.

    Examples:
        ```python
        client = Client(api_token="a0c9f3bbf7e8bc175494fc44bfc6f89aae3ebcd0", "https://app.picsellia.com")
        ```

    Attributes:
        organization: current connected (Organization)
        datalake: current connected (Datalake)
    """
    def __init__(self, api_token: str = None, organization: str = None, host: str = "https://app.picsellia.com"):
        """Initialize a connexion to Picsellia platform.

        Arguments:
            api_token (uuid4): Your API token accessible in Profile Page.
                               Defaults to None, Client will try to read environment variable PICSELLIA_TOKEN.
            organization (str, optional): Specify an organization.
                                          Defaults to None, you will be connected to your main Organization.
            host (str, optional): Define a custom host used for platform.
                                  Defaults to "https://app.picsellia.com".

        Raises:
            exceptions.NetworkError: If Picsellia platform is not responding
        """
        if api_token is None:
            if "PICSELLIA_TOKEN" in os.environ:
                token = os.environ["PICSELLIA_TOKEN"]
            else:
                raise Exception(
                    "Please set up the PICSELLIA_TOKEN environement variable or specify your token")
        else:
            token = api_token

        self.connexion = Connexion(host, token)

        # Ping platform to get username and version matching api_token
        try:
            ping_response = self.connexion.get('/sdk/v1/ping').json()
        except Exception as e:
            raise PicselliaError("Cannot connect to the platform. Please check api_token, organization and host given.\n Error is : {}".format(str(e)))

        sdk_version = ping_response["sdk_version"]

        if sdk_version != picsellia.__version__:
            logger.warning("\033[93mYou are using an outdated version of the picsellia package ({})\033[0m"
                           .format(picsellia.__version__))
            logger.warning("\033[93mPlease consider upgrading to {} with pip install picsellia --upgrade\033[0m"
                           .format(sdk_version))

        # Ask organization to platform
        if organization is not None:
            organization_response = self.connexion.get(
                '/sdk/v1/organization?name={}'.format(organization)).json()
        else:
            organization_response = self.connexion.get('/sdk/v1/organization').json()
        self.organization = Organization(
            self.connexion, organization_response["organization_id"], organization_response["organization_name"])
        self.datalake = Datalake(
            self.connexion, organization_response["datalake_id"])

        organization_name = "your" if ping_response["username"] == self.organization.name else self.organization.name + "'s"
        logger.info("Hi {}{}{}, welcome back. 🥑\nWorkspace: {}{}{} organization.".format(bcolors.OKBLUE,
                        ping_response["username"], bcolors.ENDC, bcolors.YELLOW, organization_name, bcolors.ENDC))
 
    def __str__(self) -> str:
        return "Client initialized for organization `{}`".format(self.organization.name)

    @exception_handler
    @beartype
    def get_organization(self,) -> Organization:
        """Retrieve organization information currently connected

        Examples:
            ```python
            organization = client.get_organization()
            ```

        Returns:
            The (Organization) of the client that you are using
        """
        return self.organization

    @exception_handler
    @beartype
    def get_datalake(self,) -> Datalake:
        """Retrieve datalake information of organization currently connected

        Examples:
            ```python
            datalake = client.get_datalake()
            ```

        Returns:
            The (Datalake) of the client that you are using
        """
        return self.datalake

    @exception_handler
    @beartype
    def _convert_response_to_dataset(self, response):
        dataset_id = response['dataset_id']
        dataset_name = response['dataset_name']
        dataset_version = response['version']
        return Dataset(self.connexion, dataset_id, dataset_name, dataset_version)
        
    @exception_handler
    @beartype
    def create_dataset(self, name: str, data: Union[Data, List[Data], MultiData], version: str = 'first',
                       description: str = '', private: bool = True, nb_threads: int = 20) -> Dataset:
        """Create a (Dataset) in this organization.

        This methods allows user to create a dataset into the organization currently connected.
        A dataset takes (Data) coming from organization's (Datalake) and transform it as annotable (Picture).
        User can specify name of the first version, a description and if the dataset is private or not.

        Examples:
            Create a dataset named datatest with data from datalake
            ```
                data = datalake.fetch_data()
                ds = client.create_dataset('datatest', data)
            ```

        Arguments:
            name (str): Name of the dataset. It must be unique in the organization.
            data ((Data) or (MultiData)): A bunch of data to be added to the dataset
            version (str, optional): Name of the first version. Defaults to 'first'.
            description (str, optional): A description of the dataset. Defaults to ''.
            private (bool, optional): Specify if the dataset is private. Defaults to True.

        Returns:
            A (Dataset) that you can manipulate, connected to Picsellia
        """
        assert name != '', 'Dataset name can\'t be empty'

        if isinstance(data, Data):
            data_ids = [data.id]
        else:
            data_ids = [data.id for data in data]

        assert data_ids != [], 'Please specify the assets to add to dataset'

        first_page = data_ids[:20]
        payload = json.dumps({
            'name': name,
            'version': version,
            'data_ids': first_page,
            'description': description,
            'private': private,
        })
        r = self.connexion.post(
            '/sdk/v1/organization/{}/dataset'.format(self.organization.id), data=payload).json()
        created_dataset = self._convert_response_to_dataset(r["dataset"])

        # Send other pictures
        if len(data_ids) > 20:
            other_ids = data_ids[20:]

            logger.info("🌐 Dataset is created with 20 pictures ..")
            logger.info("🌐 Adding other pictures to dataset ..")

            f = partial(mlt.mlt_add_data_to_dataset, self.connexion, created_dataset.id)
            mlt.do_multiprocess_things(f, other_ids, nb_threads)

        logger.info("📚 Dataset {} created\n📊 Size: {} pictures\n🌐 Platform url: {}"
                    .format(name, len(data_ids), created_dataset.get_resource_url_on_platform()))
        return created_dataset

    @exception_handler
    @beartype
    def get_dataset(self, name: str, version: str = "latest") -> Dataset:
        """Get a dataset by its name and version

        Examples:
            ```python
            dataset = client.get_dataset('datatest', 'first')
            ```

        Arguments:
            name (str): Name of the dataset
            version (str, optional): Version of the dataset. Defaults to "latest".

        Returns:
            A (Dataset) that you can use and manipulate
        """
        r = self.connexion.get('/sdk/v1/organization/{}/dataset/{}/{}'.format(self.organization.id, name, version)).json()
        return self._convert_response_to_dataset(r["dataset"])

    @exception_handler
    @beartype
    def get_dataset_by_id(self, id: str) -> Dataset:
        """Get a dataset by its id

        Examples:
            ```python
            dataset = client.get_dataset('918351d2-3e96-4970-bb3b-420f33ded895')
            ```

        Arguments:
            id (str): id of the dataset to retrieve

        Returns:
            A (Dataset) that you can use and manipulate
        """
        r = self.connexion.get('/sdk/v1/dataset/{}'.format(id)).json()
        return self._convert_response_to_dataset(r["dataset"])

    @exception_handler
    @beartype
    def list_datasets(self,) -> List[Dataset]:
        """Retrieve all dataset of current organization

        Examples:
            ```python
            datasets = client.list_datasets()
            ```

        Returns:
            A list of (Dataset) object that belongs to your organization
        """
        r = self.connexion.get(
            '/sdk/v1/organization/{}/datasets'.format(self.organization.id)).json()
        return list(map(self._convert_response_to_dataset, r['datasets']))

    @exception_handler
    @beartype
    def search_datasets(self, name: str = None, version: str = None) -> List[Dataset]:
        """Retrieve all dataset of current organization.

        Specifying name and/or version allows user to filter results.

        Examples:
            ```python
            datasets = client.search_datasets()
            datasets = client.search_datasets(name="datatest")
            datasets = client.search_datasets(version="latest")
            datasets = client.search_datasets(name="datatest", version="latest")
            ```

        Returns:
            A list of (Dataset) objects that belongs to your organization, with given name and/or version
        """
        params = {}
        if name is not None:
            params["name"] = name
        if version is not None:
            params["version"] = version
        r = self.connexion.get(
            '/sdk/v1/organization/{}/datasets'.format(self.organization.id), params=params).json()
        return list(map(self._convert_response_to_dataset, r['datasets']))

    @exception_handler
    @beartype
    def _convert_response_to_model(self, response: dict) -> Model:
        return Model(
            self.connexion,
            response["model_id"],
            response["network_name"],
            response["type"],
        )

    @exception_handler
    @beartype
    def create_model(self, name: str, type: str) -> Model:
        """Creates a new model.

        Arguments:
            name (str): Model name to create.
            type (str): Model type (classification, detection, segmentation).

        Returns:
            A (Model) object that you can manipulate
        """
        data = json.dumps({
            "name": name,
            "type": type
        })
        r = self.connexion.post(
            '/sdk/v1/organization/{}/model'.format(self.organization.id), data=data).json()
        created_model = self._convert_response_to_model(r["model"])
        logger.info("📚 Model {} created\n📊 Type: {}\n🌐 Platform url: {}".format(name, type, created_model.get_resource_url_on_platform()))
        return created_model

    @exception_handler
    @beartype
    def get_model(self, name: str) -> Model:
        """Retrieve a model by its name.

        Examples:
            ```python
                model = client.get_model("foo_model")
            ```
        Arguments:
            name (str): name of the model your are looking for

        Returns:
            A (Model) object that you can manipulate
        """
        params = {"name": name}
        r = self.connexion.get('/sdk/v1/organization/{}/model/search'.format(self.organization.id), params=params).json()
        return self._convert_response_to_model(r["model"])

    @exception_handler
    @beartype
    def get_model_by_id(self, id: str) -> Model:
        """Retrieve a model by its id

        Examples:
            ```python
            model = client.get_model_by_id("d8fae655-5c34-4a0a-a59a-e49c89f20998")
            ```
        Arguments:
            id (str): id of the model that you are looking for

        Returns:
            A (Model) object that you can manipulate
        """
        r = self.connexion.get('/sdk/v1/model/{}'.format(id)).json()
        return self._convert_response_to_model(r["model"])

    @exception_handler
    @beartype
    def list_models(self,) -> List[Model]:
        """List all models stored in this organization

        This will return all the models stored
        If no project is found, will throw a ResourceNotFoundError

        Examples:
            ```python
            models = client.list_models()
            ```

        Returns:
            A list of all (Model) that belong to this organization
        """
        r = self.connexion.get('/sdk/v1/organization/{}/models'.format(self.organization.id)).json()
        return list(map(self._convert_response_to_model, r["models"]))

    @exception_handler
    @beartype
    def _convert_response_to_project(self, response: dict) -> Project:
        return Project(
            self.connexion,
            response["project_id"],
            response["project_name"]
        )

    @exception_handler
    @beartype
    def create_project(self, name: str, description: str = None, dataset: Dataset = None) -> Project:
        """Create a project with given name and parameters

        This project will be registered into used organization.
        You can specify this kwargs to build project :
            - description

        Examples:
            ```python
                foo_dataset = client.get_dataset("foo", "v1") 
                my_project = client.create_project("my_project", description="My first project!", foo_dataset)
            ```
        Arguments:
            name (str): name of the project
            description (str): description of the project
            dataset (Dataset): dataset attached to this project

        Returns:
            A (Project) that you can manipulate to run experiments, or attach dataset
        """
        data = {
            "name": name
        }
        if description != None:
            data["description"] = description

        if dataset != None:
            data["dataset_id"] = dataset.id

        r = self.connexion.post(
            '/sdk/v1/organization/{}/project'.format(self.organization.id), data=json.dumps(data)).json()
        created_project = self._convert_response_to_project(r["project"])
        logger.info("📚 Project {} created\n📊 Description: {}\n🌐 Platform url: {}".format(name, description if description is not None else '', created_project.get_resource_url_on_platform()))
        return created_project

    @exception_handler
    @beartype
    def get_project(self, project_name: str) -> Project:
        """Get a project from its name

        Retrieve a project from its name.
        Project must belong to used organization.
        If no project is found, will throw a ResourceNotFoundError

        Examples:
            ```python
                my_project = client.get_project("my_project")
            ```
        Arguments:
            project_name (str): name of the project to retrieve

        Returns:
            A (Project) of your organization, you can manipulate to run experiments, or attach dataset
        """
        r = self.connexion.get('/sdk/v1/organization/{}/project'.format(
            self.organization.id), params={"name": project_name}).json()
        return self._convert_response_to_project(r["project"])

    @exception_handler
    @beartype
    def get_project_by_id(self, project_id: str) -> Project:
        """Get a project from its id

        Retrieve a project from its id.
        Project must belong to used organization.
        If no project is found, will throw a ResourceNotFoundError

        Examples:
            ```python
                my_project = client.get_project("2214aacc-b884-41e1-b70f-420c0cd7eefb")
            ```
        Arguments:
            project_id (str): id of the project to retrieve

        Returns:
            A (Project) of your organization, you can manipulate to run experiments, or attach dataset
        """
        r = self.connexion.get('/sdk/v1/project/{}'.format(project_id)).json()
        return self._convert_response_to_project(r["project"])

    @exception_handler
    @beartype
    def list_projects(self,) -> List[Project]:
        """List all projects of your organization.

        Retrieve all projects of your organization

        Examples:
            ```python
                projects = client.list_projects()
            ```

        Returns:
            A list of Project of your organization
        """
        r = self.connexion.get(
            '/sdk/v1/organization/{}/projects'.format(self.organization.id)).json()
        return list(map(self._convert_response_to_project, r['projects']))


    @exception_handler
    @beartype
    def _convert_response_to_deployment(self, response: dict) -> Deployment:
        return Deployment(
            self.connexion,
            response["pk"],
            response["name"],
            response["oracle_url"] if "oracle_url" in response else None,
            response["serving_url"] if "serving_url" in response else None
        )

    @exception_handler
    @beartype
    def get_deployment(self, name: str) -> Deployment:
        """Get a (Deployment) from its name.

        Examples:
            ```python
                deployment = client.get_deployment(
                    name="awesome-deploy"
                )
            ```
        Arguments:
            name (str): auto-generated name of your deployment.

        Returns:
            A (Deployment) object connected and authenticated to all the services.
        """
        r = self.connexion.get('/sdk/v1/organization/{}/deployment_by_name/search?name={}'.format(
            self.organization.id, name)).json()
        return self._convert_response_to_deployment(r["deployment"])


    @exception_handler
    @beartype
    def get_deployment_by_id(self, id: str) -> Deployment:
        """Get a (Deployment) from its name.

        Examples:
            ```python
                deployment = client.get_deployment_id(
                    id="YOUR DEPLOYMENT ID"
                )
            ```
        Arguments:
            id (str): deployment id displayed in your deployment settings.

        Returns:
            A (Deployment) object connected and authenticated to all the services.
        """
        r = self.connexion.get("/sdk/v1/deployment/{}".format(id)).json()
        return self._convert_response_to_deployment(r["deployment"])

    @exception_handler
    @beartype
    def list_deployments(self,) -> List[Deployment]:
        """List all (Deployment) of your organization

        Examples:
            ```python
                our_deployments = client.list_deployments()
            ```
        Arguments:
            id (str): deployment id displayed in your deployment settings.

        Returns:
            List of (Deployment): all deployments object connected and authenticated to all the services.
        """
        r = self.connexion.get("/sdk/v1/organization/{}/deployments".format(self.organization.id)).json()
        return list(map(self._convert_response_to_deployment, r['deployments']))




class PicselliaError(Exception):
    """Base class for exceptions."""
    def __init__(self, message):
        """
        Arguments:
            message (str): Informative message about the exception.
            cause (Exception): The cause of the exception (an Exception
                raised by Python or another library). Optional.
        """
        super().__init__(message)
        self.message = message

    def __str__(self):
        return self.message


class AuthenticationError(PicselliaError):
    """Raised when your token does not match to any known token"""
    pass


class UnauthorizedError(PicselliaError):
    """Raised when your token does not match to any known token"""
    pass


class ForbiddenError(PicselliaError):
    """Raised when your token does not match to any known token"""
    pass


class ResourceNotFoundError(PicselliaError):
    """Exception raised when a given resource is not found. """
    pass


class ResourceConflictError(PicselliaError):
    """Exception raised when a given resource already exists. """
    pass


class InvalidQueryError(PicselliaError):
    """ Indicates a malconstructed or unsupported query. This can be the result of either client
    or server side query validation. """
    pass


class NetworkError(PicselliaError):
    """Raised when an HTTPError occurs."""
    pass


class ApiLimitError(PicselliaError):
    """ Raised when the user performs too many requests in a short period
    of time. """
    pass


class ProcessingError(PicselliaError):
    """Raised when an algorithmic error occurs."""
    pass


class TyperError(PicselliaError):
    """Raised when a method argument has the wrong type"""
    pass


class UnsufficientRessourcesError(PicselliaError):
    """Raised when your token does not match to any known token"""
    pass


class NoDataError(PicselliaError):
    """Raised when you try to retrieve data from an empty datalake"""
    pass

class ImpossibleAction(PicselliaError):
    """Raised when an action is impossible"""
    pass

class PicselliaOutError(Exception):
    def __init__(self, picsellia_error : PicselliaError) -> None:
        super().__init__(picsellia_error)
        self.message = picsellia_error.message

    def __str__(self) -> str:
        return "Something went wrong.\n{}.".format(self.message)



from itertools import chain
import json
import logging
from multiprocessing import Pool, Value
from typing import Callable, List, Union
from uuid import uuid4
import requests
import os
from picsellia.exceptions import ResourceNotFoundError
from picsellia.sdk.connexion import Connexion
from picsellia.sdk.data import Data, MultiData
from picsellia.sdk.picture import MultiPicture, Picture
from picsellia.utils import print_next_bar
from PIL import Image
import warnings
from beartype.roar import BeartypeDecorHintPep585DeprecationWarning
warnings.filterwarnings("ignore", category=BeartypeDecorHintPep585DeprecationWarning)
logger = logging.getLogger('picsellia')


def chunks(lst, n):
    """Yield successive n-sized chunks from lst."""
    for i in range(0, len(lst), n):
        yield lst[i:i + n]

# Multiprocessed init
def pool_init(length, counter):
    global mlt_counter
    global mlt_total_length
    global mlt_data_list

    mlt_counter = counter
    mlt_total_length = length
    mlt_data_list = []


def do_multiprocess_things(f : Callable, data_list : list, nb_threads : int = 20) -> list:
    nb_threads = min(len(data_list), nb_threads)
    chunk_size = (len(data_list) // nb_threads) + 1
    infos_split = list(chunks(data_list, chunk_size))
    counter = Value('i', 0)
    print_next_bar(0, len(data_list))
    with Pool(nb_threads, initializer=pool_init, initargs=(len(data_list), counter,)) as p:
        result = p.map(f, infos_split)

    return list(chain.from_iterable(result))

def mlt_download_list_external_picture(png_dir : str, infos : list) -> list:
    global mlt_counter
    global mlt_total_length

    downloaded = []
    for info in infos:
        pic_name = os.path.join(png_dir, os.path.split(info['external_picture_url'])[-1])
        if not os.path.isfile(pic_name) and _download_external_picture(info=info, pic_name=pic_name):
            downloaded.append(pic_name)

    with mlt_counter.get_lock():
        mlt_counter.value += len(infos)
        print_next_bar(mlt_counter.value, mlt_total_length)

    return downloaded

def _download_external_picture(info, pic_name):
    global mlt_counter
    global mlt_total_length
    try:
        response = requests.get(info["signed_url"], stream=True)
        if response.status_code == "404":
            raise ResourceNotFoundError('Picture does not exist on our server.')
        with open(pic_name, 'wb') as handler:
            for data in response.iter_content(chunk_size=1024):
                handler.write(data)
    except Exception:
        logger.error(f"Image {pic_name} can't be downloaded")
        return False
    return True

def mlt_download_list_data_or_pics(connexion : Connexion, lake_or_set_id : str, target_path: str, some_data_or_pics : Union[Picture, MultiPicture, Data, MultiData]) -> Union[MultiPicture, MultiData]:
    global mlt_counter
    global mlt_total_length

    if isinstance(some_data_or_pics, Data):
        some_data_or_pics = MultiData(connexion, lake_or_set_id, [some_data_or_pics])
    if isinstance(some_data_or_pics, Picture):
        some_data_or_pics = MultiPicture(connexion, lake_or_set_id, [some_data_or_pics])

    for pic in some_data_or_pics:
        path = os.path.join(target_path, pic.external_url)
        connexion.download_some_file(False, pic.internal_key, path, False)

    with mlt_counter.get_lock():
        mlt_counter.value += len(some_data_or_pics)
        print_next_bar(mlt_counter.value, mlt_total_length)

    return some_data_or_pics

def mlt_add_data_to_dataset(connexion : Connexion, dataset_id : str, data_ids: List[str]) -> list:
    """Add retrieved data from datalake to this dataset

    It will add given data into this dataset object.

    Examples:
        ```python
            data = datalake.fetch_data()[:10]
            bar_dataset.add_pictures(data)
        ```
    Arguments:
        data ((Data), List[(Data)] or (MultiData)): data to add to dataset
    """
    global mlt_counter
    global mlt_total_length

    payload = json.dumps({
        'picture_ids': data_ids
    })
    connexion.post('/sdk/v1/dataset/{}/pictures'.format(dataset_id), data=payload)
    with mlt_counter.get_lock():
        mlt_counter.value += len(data_ids)
        print_next_bar(mlt_counter.value, mlt_total_length)

    return data_ids

def mlt_upload_data(connexion : Connexion, datalake_id : str, tags: List[str], source: str, paths: List[str]):
    global mlt_data_list
    global mlt_total_length
    global mlt_counter

    data_list = []
    for path in paths:
        try:
            external_url = os.path.split(path)[-1]
            internal_key = os.path.join(str(uuid4())) + '.' + external_url.split('.')[-1]
            with Image.open(path) as image:
                width, height = image.size
            connexion.push_to_s3(path, internal_key)
            data = json.dumps({
                'internal_key': internal_key,
                'external_url': external_url,
                'height': height,
                'width': width,
                'tags': tags,
                'source': source
            })
            r = connexion.post('/sdk/v1/datalake/{}/data/add'.format(datalake_id), data=data).json()
            data_list.append(Data(connexion, datalake_id, r["data"]["picture_id"], r["data"]["external_url"], r["data"]["internal_key"]))
        except Exception as e:
            logger.error('\nCould not upload {} because {}'.format(path, str(e)))
            continue
    with mlt_counter.get_lock():
        mlt_counter.value += len(paths)
        print_next_bar(mlt_counter.value, mlt_total_length)
    mlt_data_list += data_list
    return mlt_data_list
        

def init_pool_annotations(p, nb, req, ds_id, p_size):
    global page_done
    global nb_pages
    global connexion
    global annotation_list
    global dataset_id
    global page_size
    page_done = p
    nb_pages = nb - 1
    connexion = req
    dataset_id = ds_id
    annotation_list = []
    page_size = p_size


def dl_annotations(page_list):
    global page_done
    global nb_pages
    global connexion
    global annotation_list
    global dataset_id
    global page_size

    annotation_list = []
    for page in page_list:
        params = {
            'limit': page_size,
            'offset': page_size*page,
            'snapshot': False
        }
        logging.error("coucou")
        r = connexion.get(
            '/sdk/v1/dataset/{}/annotations'.format(dataset_id), params=params)
        annotation_list += r.json()["annotations"]
        with page_done.get_lock():
            page_done.value += 1
        print_next_bar(page_done.value, nb_pages)
    return annotation_list



from beartype.roar._roarexc import (
    BeartypeCallHintPepParamException,
)
from functools import wraps
import time
from picsellia import exceptions


def exception_handler(func):
    def inner(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except BeartypeCallHintPepParamException as exc:
            exc = str(exc).split('@beartyped ')
            message = exc[1]
            raise TypeError(message)
    return inner


def retry(exceptions, total_tries=4, initial_wait=0.5, backoff_factor=2):
    """
    calling the decorated function applying an exponential backoff.
    Arguments:
        exceptions: Exeption(s) that trigger a retry, can be a tuple
        total_tries: Total tries
        initial_wait: Time to first retry
        backoff_factor: Backoff multiplier (e.g. value of 2 will double the delay each retry).
    """
    def retry_decorator(f):
        @wraps(f)
        def func_with_retries(*args, **kwargs):
            _tries, _delay = total_tries + 1, initial_wait
            while _tries > 1:
                try:
                    return f(*args, **kwargs)
                except exceptions:
                    _tries -= 1
                    if _tries == 1:
                        raise
                    time.sleep(_delay)
                    _delay *= backoff_factor

        return func_with_retries
    return retry_decorator




from typing import List
from picsellia.types.prediction_format import PredictionFormat


class SegmentationPredictionFormat(PredictionFormat):

    def __init__(self, label_ids : List[int], boxes : List[List[int]], detection_scores : List[float]):
        self.label_ids = label_ids
        self.boxes = boxes
        self.detection_scores = detection_scores
        self.check_validity()

    def check_validity(self) -> bool:
        assert isinstance(self.label_ids, List[int]), "Labels shall be a list of integer. Each one represening a label in labelmap"
        assert isinstance(self.boxes, List[List[int]]), "Boxes shall be a List of List of integer values."
        assert isinstance(self.detection_scores, List[int]), "Scores shall be a List of float values."

        for box in self.boxes:
            assert len(box) > 0, "One box is empty"

        for score in self.detection_scores:
            assert score >= 0, "Score shall be positive values"

    def to_payload(self) -> dict:
        self.check_validity()
        return {
            "labels": self.label_ids,
            "detection_scores": self.detection_scores,
            "boxes": self.boxes
        }


__version__ = "5.1.10"

import sys
import os
from setuptools import find_packages
import picsellia
from picsellia.client import Client  # Not used here but allows to : from picsellia import Client directly
import logging
import logging.config
for p in find_packages(where=picsellia.__path__[0]):
    sys.path.append(os.path.join(picsellia.__path__[0], p.replace('.', '/')))

sys.path.append(picsellia.__path__[0])


logger = logging.getLogger('picsellia')

logger.addHandler(logging.NullHandler())

try:
    custom_logging = os.environ["PICSELLIA_SDK_CUSTOM_LOGGING"]
except KeyError:
    custom_logging = False


if not custom_logging:
    try:
        logging.config.fileConfig('{}/conf/default_logging.conf'.format(picsellia.__path__[0]))
    except Exception as e:
        print('Error while loading conf file for logging. No logging done.')
